/*!
	\file Contour2D.h
	\brief Contour2D class declaration

	Information :
		- Author	: Vincent PRAT (vprat@ifrance.com)
		- Created	: 10-07-2001 (dd-mm-yyyy)
		- Modified	: 12-07-2001 (dd-mm-yyyy)

	Copyright  2001 Vincent PRAT
*/

#ifndef __CONTOUR2D_H__
#define __CONTOUR2D_H__

#include <math.h>			// sqrt
#include "VectorNd.h"		// Vector2d, Vector3d and Vector4d
#include "NURBSSurface.h"	// TrimmedNURBSSurface

/*!
	\brief Default number of vertices generated for one curve
*/
#define DEFAULT_TESSELATION_VALUE 20

/*!
	\brief A Contour2D is a NURBS curve.

	Contour2D is used by a trimmed surface to represent holes in the surface.
	Compared to a classic NURBS curve, we added an array to save the points of the
	surface corresponding to the points of the contour and a few useful functions 
	to detect if a point is in the contour or to find the closest point of the contour.
*/
class Contour2D
{
	// Not really a nice thing but so handy for the constructor of
	// TrimmedNURBSSurface !
	friend class TrimmedNURBSSurface;

protected:
	/*!
		\brief order of the NURBS curve

		This integer must be strictly higher than 1 (>=2)
	*/
	int	nOrder;

	/*!
		\brief Number of control points defining the curve
	*/
	int	nNumberOfControlPoints;		
	
	/*!
		\brief Number of values in the knot span

		This is deduced from other constants :
		numberOfKnots = numberOfControlPoints + order
	*/
	int	nNumberOfKnots;				
	
	/*!
		\brief Current tesselation value of the curve

		This represents the number of segments that will approximate the
		mathematical curve. Thus, there are tesselation+1 vertices.
	*/
	int	nTesselation;

	/*!
		\brief Degree of the curve

		This is deduced from the order : degree = order-1
	*/
	int	nDegree;

	/*!
		\brief Number of valid knot spans in the knot vector

		This is deduced from other constants :
		numberOfKnotSpans = NumberOfKnots - 2*degree
	*/
	int	nNumberOfKnotSpans;

	/*!
		\brief Top-right point of the bounding box
	*/
	Vector2d	BBMax;				

	/*!
		\brief Lower-left point of the bounding box
	*/
	Vector2d	BBMin;

	/*!
		\brief pointer to coefficients of the basis function polynomials

		These values are pre-calculated at surface creation only. 
		The size of the array is order*order*numberOfKnotSpans
	*/
	float*	pfBasisCoefficients;

	/*!
		\brief pointer to values of the basis function at the tesselated points

		These values are pre-calculated each time the tesselation value changes.
		The size of the array is tesselation+1
	*/
	float*	pfBasisValues;

	/*!
		\brief pointer to knot vector

		The size of the array is numberOfKnots
	*/
	float*	pfKnots;

	/*!
		\brief Pointer to the knot span corresponding to a given tesselated point

		The size of the array is tesselation+1
	*/
	int*	pnTessKnotSpan;		
	
	/*!
		\brief Pointer to the array of control points

		The size of the array is numberOfControlPoints
	*/
	Vector4d*	pControlPoints;	

	/*!
		\brief Pointer to precalculated surface points

		These points are dependant on the surface trimmed by this contour.
		We precalculate them to avoid costly computation when trimming the surface.
		The size of the array is tesselation+1
	*/
	Vector3d*	pSurfacePoints;		

	/*!
		\brief Pointer to tesselated contour points

		Those points are in the surface's parametric plane.
		The size of the array is tesselation+1
	*/
	Vector2d*	pPoints;			
	
	/*!
		\brief determine whether a point is in the bounding box of this contour.
		\param u u coordinate of the point to test
		\param v v coordinate of the point to test
		\return true if the point is in the bounding box, false if not
	*/
	bool IsInBB( float u, float v );

	/*!
		\brief Tesselates the surface

		Basically, this function fills the array pPoints. It uses the precalculated
		values of the basis functions.
	*/
	void Tesselate();

	/*!
		\brief Initialises and computes every variable that is not provided during construction.

		This calculates for instance the degree, the number of knots and such deduced 
		values. Also pre-calculates basis coefficients, basis values and launches the 
		first tesselation.
	*/
	void Initialise();

	/*!
		\brief Fills the pfBasisCoefficients array.

		This function uses the ComputeCoefficient function to find the particular value
		of a certain coefficient. Function called only once by Initialise().
	*/
	void EvaluateBasisFunctionsCoefficients();

	/*!
		\brief Fills the pfBasisValues array

		This function calculates the basis values for each tesselated point. At the
		same time, we also fill the pnTessKnotSpan array. We use the formula :
		\image html formula02.gif
	*/
	void EvaluateBasisFunctions();

	/*!
		\brief Computes one particular polynomial coefficient (which depends only on the knotspan we are in)
		\param  nInterval knot vector interval for this particular coefficient
		\param i see formula
		\param p see formula
		\param k see formula
		This function is recursive and uses the formula :
		\image html formula01.gif
	*/
	float ComputeCoefficient( int nInterval, int i, int p, int k );

public:
	/*!
		\brief Default constructor (does nothing)
	*/	
	Contour2D() {}

	/*!
		\brief Constructor of a Contour2D object
		\param _nOrder order of the NURBS curve
		\param _nNumberOfControlPoints number of control points defining the curve
		\param _pControlPoints Pointer to the array of control points
		\param _pfKnots Pointer to the knot vector

		This constructor will clone the knots and the control points and the call Initialise().
	*/	
	Contour2D( 
		int _nOrder, 
		int _nNumberOfControlPoints, Vector4d *_pControlPoints, 
		float* _pfKnots );

	/*!
		\brief Destructor of the object

		Frees all the memory that was allocated dynamically
	*/	
	virtual ~Contour2D();
	
	/*!
		\brief Test whether a point is inside the curve
		\param u u coordinate of the point to test
		\param v v coordinate of the point to test
		\return true if the point is in the curve, false if not

		First checks if the point is in the bounding box and in that case, 
		uses the ray-crossing method to determine if the point is actually 
		inside the curve.

		See comp.graphics.algorithms FAQ, subject 2.03 for more information 
		about this method
	*/	
	bool IsInside( float u, float v );

	/*!
		\brief finds the closest element of pPoints to point (u,v)
		\param u u coordinate of the given point
		\param v v coordinate of the given point
		\param uMin u minimum value (to avoid to go out of the knot vector)
		\param uMax u maximum value (to avoid to go out of the knot vector)
		\param vMin v minimum value (to avoid to go out of the knot vector)
		\param vMax v maximum value (to avoid to go out of the knot vector)
		\return the index of the closest point in pPoints

	*/	
	int FindNearestPoint( 
		float u, float v, 
		float uMin, float uMax, 
		float vMin, float vMax );

	/*!
		\brief Computes the coordinates of the surface points corresponding to the points of pPoints
		\param surface Pointer to the surface corresponding to those points

		We fill the array pSurfacePoints. In order to do this, we have to calculate the value of
		the surface's basis function for the values of the parameters given in pPoints.
	*/	
	void ComputeSurfacePoints( TrimmedNURBSSurface *surface );
};

/*!
	\brief shortcut to design a pointer to a Contour2D object.
*/
typedef Contour2D *LPCONTOUR2D;

#endif